/* 
 * Copyright (c) 2010 Michael J. Wilson
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

// Michael J. Wilson
// Music 256a Fall 2010
// Final Project

#include "Instrument.h"
#include "Utility.h"
#include <cmath>
#include <iostream>

Instrument::~Instrument()
{
    m_mode = UNIT_SINE_OFF_MODE;
    free_sms_data();
}

Instrument::Instrument()
{
    m_frequency = 0;
    m_current_sample = 0;
    m_state = UNIT_SINE_OFF;
    m_gain = 0.0;
    m_phase = 0.0;
    m_mode = UNIT_SINE_SINE_MODE;
    synth_is_initialized = false;
}

void Instrument::play(int note)
{
    // Add note offset; for now just pick an arbitrary default value.
    m_frequency = Utility::midi_to_frequency(note + 62);
    m_current_sample = 0;
    m_sample_index = 0;
    m_sms_buffer_index = 0;
    m_state = UNIT_SINE_ATTACK;

    // If we're in SMS mode, synthesize the next sample:
    if (m_mode == UNIT_SINE_SMS_MODE)
    {
        Utility::get_utility()->lock();
        synthesize_note(note);
        Utility::get_utility()->unlock();
    }
}

void Instrument::init_synth()
{
    // Initialization
    sms_initSynthParams(&synthParams);
    sms_initSynth(&smsHeader, &synthParams);
    sms_init();
    synth_is_initialized = true;
}

void Instrument::free_synth()
{
    if (synth_is_initialized)
    {
        sms_freeSynth(&synthParams);
        sms_free();
    }
    synth_is_initialized = false;
}

void Instrument::synthesize_note(int note)
{
    // Code adapted from smsSynth.c which is Copyright (c) 2008 MUSIC
    // TECHNOLOGY GROUP (MTG) UNIVERSITAT POMPEU FABRA

    // Bookkeeping on datastructures
    m_sms_buffer.clear();

    // Synthesize sample by checking beat and pitch from current note
    // (set timescale and pitch factors accordingly)
    SMS_Data smsFrameL;
    SMS_Data smsFrameR;
    SMS_Data smsFrame;
    float *pFSynthesis;
    long iSample;
    long nSamples;
    long iLeftFrame;
    long iRightFrame;
    float fFrameLoc;
    float fFsRatio;
    float fLocIncr;
    float timeFactor = 1.0;
    int i;

    // Initialize
    init_synth();

    // Turn on transposing and res gain
    synthParams.modParams.doTranspose = 1;
    synthParams.modParams.doResGain = 1;

    // Setup for synthesis
    sms_allocFrameH(&smsHeader, &smsFrameL);
    sms_allocFrameH(&smsHeader, &smsFrameR);
    sms_allocFrameH(&smsHeader, &smsFrame);

    // Set timescales and transposes:
    long recorded_samples = m_sample.size();
    long samples_per_beat = ((Utility::sample_rate) * (Utility::note_interval) / 1000);
    timeFactor = (float)samples_per_beat / (float)recorded_samples;
    synthParams.modParams.transpose = note;

    if ((pFSynthesis = (float *) calloc(synthParams.sizeHop, sizeof(float))) == NULL)
    {
        std::cout << "Could not allocate memory for pFSynthesis" << std::endl;
    }

    iSample = 0;
    // Number of samples is a factor of the ratio of samplerates
    // Multiply by timeFactor to increase samples to desired file length
    // pSmsHeader->iSamplingRate is from original analysis
    fFsRatio = (float) synthParams.iSamplingRate / smsHeader.iSamplingRate;
    nSamples = smsHeader.nFrames * synthParams.origSizeHop * timeFactor * fFsRatio;

    // Divide timeFactor out to get the correct frame
    fLocIncr = smsHeader.iSamplingRate / 
        ( synthParams.origSizeHop * synthParams.iSamplingRate * timeFactor); 

    while (iSample < nSamples)
    {
        // Left and right frames around location, guarding for end of file
        fFrameLoc =  iSample *  fLocIncr;
        iLeftFrame = MIN(smsHeader.nFrames - 1, floor (fFrameLoc));
        iRightFrame = (iLeftFrame < smsHeader.nFrames - 2) ? (1+ iLeftFrame) : iLeftFrame;

        // Get frames and interpolate
        sms_copyFrame(&smsFrameL, m_sms_data[iLeftFrame]);
        sms_copyFrame(&smsFrameR, m_sms_data[iRightFrame]);
        sms_interpolateFrames (&smsFrameL, &smsFrameR, &smsFrame,
                               fFrameLoc - iLeftFrame);

        // Do the resynthesis
        sms_modify(&smsFrame, &synthParams.modParams); 
        sms_synthesize(&smsFrame, pFSynthesis, &synthParams);

        // Put sound in buffer
        for (i = 0; i < synthParams.sizeHop; i++)
        { 
            m_sms_buffer.push_back(pFSynthesis[i]);
        }
    
        iSample += synthParams.sizeHop;
    }

    // Cleanup
    sms_freeFrame(&smsFrameL);
    sms_freeFrame(&smsFrameR);
    sms_freeFrame(&smsFrame);
    free(pFSynthesis);
    free_synth();
}

void Instrument::stop()
{
    m_state = UNIT_SINE_RELEASE;
}

float Instrument::get_next_sample()
{
    float result = 0.0;
    long samples_per_beat = ((Utility::sample_rate) * (Utility::note_interval) / 1000);

    switch(m_state)
    {
    case UNIT_SINE_ATTACK:
    case UNIT_SINE_SUSTAIN:
        // Put a decay on:
        m_current_sample++;
        if(m_current_sample > (samples_per_beat - (Utility::sample_rate * 10 / 1000)))
        {
            m_state = UNIT_SINE_RELEASE;
        }
        // No break
    case UNIT_SINE_RELEASE:
        if ((m_mode == UNIT_SINE_SMS_MODE) && (m_sms_buffer.size() > 0))
        { 
            // SMS mode
            result = m_gain * m_sms_buffer[m_sms_buffer_index];
            m_sms_buffer_index++;
            m_sms_buffer_index %= m_sms_buffer.size();
        }
        else if ((m_mode == UNIT_SINE_SAMPLE_MODE) && (m_sample.size() > 0))
        {
            // Sample playback mode
            result = m_gain * m_sample[m_sample_index];
            m_sample_index++;            
            m_sample_index %= m_sample.size();
        }
        else if (m_mode == UNIT_SINE_SINE_MODE)
        {
            // No sample, default to sine wave
            m_phase += (2 * Utility::PI * m_frequency / (float)Utility::sample_rate);
            while (m_phase > 2 * Utility::PI)
            {
                m_phase -= (2 * Utility::PI);
            }
            result = m_gain * std::sin(m_phase);
        }
        else if (m_mode == UNIT_SINE_OFF_MODE)
        {
            // Do nothing
        }
        break;
    case UNIT_SINE_OFF:
    default:
        break;
    }

    switch(m_state)
    {
    case UNIT_SINE_ATTACK:
        if (m_gain < 1.0)
        {
            m_gain += .001;
        }
        if (m_gain > 1.0)
        {
            m_gain = 1.0;
            m_state = UNIT_SINE_SUSTAIN;
        }
        break;
    case UNIT_SINE_RELEASE:
        if (m_gain > 0.0001)
        {
            m_gain *= .99;
        }
        if (m_gain < 0.000000001)
        {
            m_gain = 0.0;
            m_phase = 0.0;
            m_state = UNIT_SINE_OFF;
        }
        break;
    default:
        break;
    }

    return (result);
}

void Instrument::set_mode(InstrumentMode mode)
{
    if(((mode == UNIT_SINE_SAMPLE_MODE) && (m_sample.size() > 0)) ||
       ((mode == UNIT_SINE_SMS_MODE) && (m_sms_data.size() > 0)) ||
       (mode == UNIT_SINE_SINE_MODE) ||
       (mode == UNIT_SINE_OFF_MODE))
    {
        m_mode = mode;
    }
}

void Instrument::get_sample_from_utility()
{
    m_sample.clear();
    m_sample_index = 0;
    // Also might want some onset detection eventually.
    for(size_t i= 0; i < Utility::get_utility()->current_recording.size(); i++)
    {
        m_sample.push_back(Utility::get_utility()->current_recording[i]);
    }
}

void Instrument::free_sms_data()
{
    for (size_t i=0; i < m_sms_data.size(); ++i)
    {
        sms_freeFrame(m_sms_data[i]);
        delete m_sms_data[i];
    }
    m_sms_data.clear();
}

void Instrument::do_sms_processing()
{
    // Code adapted from smsAnal.c which is Copyright (c) 2008 MUSIC
    // TECHNOLOGY GROUP (MTG) UNIVERSITAT POMPEU FABRA

    // Bookkepping on datastructures
    m_sms_buffer.clear();
    m_sms_buffer_index = 0;
    free_sms_data();

    SMS_AnalParams analParams;
    SMS_SndHeader soundHeader;
    float pSoundData[SMS_MAX_WINDOW];
    int iDoAnalysis = 1;
    int iFrame = 0;
    long iStatus = 0;
    long iSample = 0;
    long sizeNewData = 0;
    long i;

    // Initialization
    soundHeader.nSamples = m_sample.size();
    soundHeader.iSamplingRate = Utility::sample_rate;
    soundHeader.channelCount = 1;
    soundHeader.iReadChannel = 0;
    soundHeader.sizeHeader = 0;
    sms_initAnalParams(&analParams);
    analParams.iSamplingRate = Utility::sample_rate;
    sms_init();
    sms_initAnalysis(&analParams, &soundHeader);
    sms_fillHeader(&smsHeader, &analParams, "");

    // Run SMS analysis on the current recording
    // Perform analysis
    while(iDoAnalysis > 0)
    {
        // Allocate output SMS record
        SMS_Data* pSmsData = new SMS_Data;
        sms_allocFrameH(&smsHeader, pSmsData);

        iSample += sizeNewData;
        if((iSample + analParams.sizeNextRead) < soundHeader.nSamples)
        {
            sizeNewData = analParams.sizeNextRead;
        }
        else
        {
            sizeNewData = soundHeader.nSamples - iSample;
        }

        // Get one frame of sound
        for (i = 0; i < sizeNewData; i++)
        {
            pSoundData[i] = m_sample[i+iSample];
        }

        // Perform analysis of one frame of sound
        iStatus = sms_analyze(sizeNewData, pSoundData, pSmsData, &analParams);

        // Store the analysis
        if (iStatus == 1)
        {
            // Store the smsData in a vector here
            m_sms_data.push_back(pSmsData);
            iFrame++;
        }
        else if (iStatus == -1)
        {
            // Get rid of the extra last frame
            sms_freeFrame(pSmsData);
            delete pSmsData;
            iDoAnalysis = 0;
            smsHeader.nFrames = iFrame;
            // Number of frames is also the length of m_sms_data
        }
    }

    // Shut down SMS
    sms_freeAnalysis(&analParams);
    sms_free();
}
