/* 
 * Copyright (c) 2010 Michael J. Wilson
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

// Michael J. Wilson
// Music 256a Fall 2010
// Final Project

#ifndef __UTILITY_H__
#define __UTILITY_H__

#include <string>
#include <vector>
#include <pthread.h>

const int DEBUG = 1;

class RtAudio;
class WidgetMain;

/// Singleton utility class, contains data that should be available to
/// the entire program
class Utility
{
public:
    /// Destructor
    ~Utility();

    /// Get the utility instance.  Lazily constructed.
    static Utility* get_utility();

    /// Ratio of circle's circumferance to its diameter
    static const float PI;

    /// Very small value for floating point comparisons
    static const float EPSILON;

    /// Sample rate
    static const unsigned int MAXIMUM_SAMPLE_RATE;

    /// Sample rate
    static unsigned int sample_rate;

    /// Number of tracks currently active
    static unsigned int num_tracks;

    /// Number of output channels
    static const unsigned int OUTPUT_CHANNELS;

    /// Name of the program, what the user would type at the command line
    static const std::string PROGRAM_NAME;

    /// A pointer to the DAC
    RtAudio* dac;

    /// Get a random number in the specified range.  No assurances on
    /// its distribution.
    float rand(float min, float max);

    /// Get the frequency of the specified MIDI note
    static float midi_to_frequency(int note);

    /// Get the closest MIDI note to the specified frequency
    static int frequency_to_midi(float frequency);

    /// Get the pitch name for the specified MIDI note
    static std::string midi_to_string(int note);

    /// The number of timesteps (horizontal axis) in the matrix editor
    static const unsigned int matrix_num_steps;

    /// The number of note values (vertical axis) in the matrix editor
    static const unsigned int matrix_num_note_values;

    /// Special flag for note off
    static const int NOTE_OFF;

    /// Special flag for note continue
    static const int NOTE_CONTINUE;

    /// Lock mutex
    void lock();

    /// Unlock mutex
    void unlock();

    /// Are we currently recording?
    /// Lock mutex before accessing
    bool is_recording();
    
    /// Start recording
    /// Lock mutex before accessing
    void start_recording();
    
    /// Stop recording
    /// Lock mutex before accessing
    void stop_recording();

    /// Consume one sample from the record buffer
    /// Lock mutex before accessing
    void consume_sample(float sample);

    /// Current recording
    /// Lock mutex before accessing
    std::vector<float> current_recording;

    /// Length of a note, in ms (set by transport widget)
    /// Lock mutex before accessing
    static int note_interval;

private:
    /// Private constructor to follow singleton pattern
    Utility();

    /// Mutex for sound operations
    pthread_mutex_t sound_mutex;

    /// Flag for if we're recording or not
    bool currently_recording;
};

#endif
