/* 
 * Copyright (c) 2010 Michael J. Wilson
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

// Michael J. Wilson
// Music 256a Fall 2010
// Final Project

#include <QPainter>
#include <QMouseEvent>
#include <cassert>

#include "Utility.h"
#include "TrackData.h"
#include "WidgetMatrixEditor.h"

const int min_scale = 15;

WidgetMatrixEditor::WidgetMatrixEditor(TrackData* track_data, QWidget* parent)
    : QWidget(parent)
{
    m_current_beat = -1;
    m_track_data = track_data;
    setBackgroundRole(QPalette::Midlight);
    setAutoFillBackground(true);
    setMinimumSize(Utility::matrix_num_steps * min_scale, Utility::matrix_num_note_values * min_scale);
}

int WidgetMatrixEditor::get_note_at(unsigned int index)
{
    assert(index < Utility::matrix_num_note_values);
    return m_track_data->m_sequence[index];
}

void WidgetMatrixEditor::paintEvent(QPaintEvent* event)
{
    unsigned int i;
    QPainter painter(this);

    // Current beat bar
    painter.setBrush(QColor(Qt::red).darker(100 + 10*m_fade_value));
    if (m_current_beat >= 0)
    {
        painter.drawRect(x_line_coord(m_current_beat),
                         0,
                         x_line_coord(m_current_beat + 1) - x_line_coord(m_current_beat),
                         height());
    }

    // Vertical subdivisions
    for (i = 1; i < Utility::matrix_num_note_values; i++)
    {
        painter.drawLine(0, y_line_coord(i),
                         width()-1, y_line_coord(i));
    }

    // Horizontal subdivisions
    for (i = 1; i < Utility::matrix_num_steps; i++)
    {
        painter.drawLine(x_line_coord(i),0,
                         x_line_coord(i),height()-1);
    }

    // Matrix state
    for (i = 0; i < Utility::matrix_num_steps; i++)
    {
        if (m_track_data->m_sequence[i] == Utility::NOTE_CONTINUE)
        {
            // TODO (NOTE_CONTINUE is unused in this version)
        }
        if (m_track_data->m_sequence[i] != Utility::NOTE_OFF)
        {
            QLinearGradient gradient(QPointF(0,0),QPointF(1,1));
            gradient.setCoordinateMode(QGradient::ObjectBoundingMode);
            gradient.setColorAt(1, Qt::blue);
            gradient.setColorAt(0, Qt::white);
            painter.setBrush(QBrush(gradient));
            painter.drawRect(get_note_bounding_box(m_track_data->m_sequence[i], i));
        }
    }
}

QRect WidgetMatrixEditor::get_note_bounding_box(unsigned int subdivision, unsigned int note)
{
    int x = x_line_coord(note);
    int y = y_line_coord(subdivision);
    int w = x_line_coord(note+1) - x;
    int h = y_line_coord(subdivision + 1) - y;

    return QRect(x,y,w,h);
}

void WidgetMatrixEditor::mouseMoveEvent(QMouseEvent *event)
{
    int note = get_note_from_point(event->pos());
    int subdivision = get_subdivision_from_point(event->pos());

    if (event->buttons() & Qt::LeftButton)
    {
        // Need to check bounds since they could click right on a boundary.
        if ((note >= 0) && (subdivision >= 0))
        {
            // TODO check if it's not a continuing note
            m_track_data->m_sequence[subdivision] = note;
        }
    }
    else if (event->buttons() & Qt::RightButton)
    {
        // Let them right-click anywhere in the column to delete the
        // note.
        if (subdivision >= 0)
        {
            // TODO check if it's not a continuing note
            m_track_data->m_sequence[subdivision] = -1;
        }
    }

    update();
}

void WidgetMatrixEditor::mousePressEvent(QMouseEvent *event)
{
    int note = get_note_from_point(event->pos());
    int subdivision = get_subdivision_from_point(event->pos());
    if (event->button() == Qt::LeftButton)
    {
        // Need to check bounds since they could click right on a boundary.
        if (note >= 0 && subdivision >= 0)
        {
            // TODO check if it's not a continuing note
            m_track_data->m_sequence[subdivision] = note;
        }
    }
    if (event->button() == Qt::RightButton)
    {
        // Let them right-click anywhere in the column to delete the
        // note.
        if (subdivision >= 0)
        {
            // TODO check if it's not a continuing note
            m_track_data->m_sequence[subdivision] = -1;
        }
    }

    // Tell widget to redraw
    update();
}

void WidgetMatrixEditor::mouseReleaseEvent(QMouseEvent *event)
{
    if (event->button() == Qt::LeftButton)
    {
        // TODO perhaps use this to extend long boxes / write continues
    }
}

int WidgetMatrixEditor::y_line_coord(unsigned int index)
{
    // Flip top-to-bottom
    return (((Utility::matrix_num_note_values - (index)) * height())/Utility::matrix_num_note_values);
}

int WidgetMatrixEditor::x_line_coord(unsigned int index)
{
    return ((index*width())/Utility::matrix_num_steps);
}

int WidgetMatrixEditor::get_note_from_point(const QPoint &pos)
{
    int result = -1;
    unsigned int i;
    for (i = 0; (result == -1) && (i < Utility::matrix_num_note_values); i++)
    {
        // Recall that y-coordinates are flipped.
        if ((pos.y() < y_line_coord(i)) &&
            (pos.y() > y_line_coord(i+1)))
        {
            result = i;
        }
    }
    return result;
}

int WidgetMatrixEditor::get_subdivision_from_point(const QPoint &pos)
{
    int result = -1;
    unsigned int i;
    for (i = 0; (result == -1) && (i < Utility::matrix_num_steps); i++)
    {
        if ((pos.x() > x_line_coord(i)) &&
            (pos.x() < x_line_coord(i+1)))
        {
            result = i;
        }
    }
    return result;
}

void WidgetMatrixEditor::set_beat(int beat)
{
    m_current_beat = beat;
    m_fade_value = 0;
    update();
}


void WidgetMatrixEditor::fade()
{
    m_fade_value++;
    update();
}
