/* 
 * Copyright (c) 2010 Michael J. Wilson
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

// Michael J. Wilson
// Music 256a Fall 2010
// Final Project

#include <QApplication>
#include <QFont>
#include <QPushButton>
#include <QWidget>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QGridLayout>
#include <QSpinBox>
#include <QLabel>
#include <QFileDialog>
#include <QCloseEvent>
#include <QtGui>
#include <QTextEdit>

#include "WidgetTrack.h"
#include "WidgetMatrixEditor.h"
#include "TrackData.h"
#include "Utility.h"
#include "Instrument.h"

#include <iostream>

WidgetTrack::~WidgetTrack()
{
    // Stop recording if we are recording and this track is being deleted
    if (m_is_recording)
    {
        Utility::get_utility()->lock();
        Utility::get_utility()->stop_recording();
        Utility::get_utility()->unlock();
    }
    delete m_track_data;
    delete m_instrument;
}

WidgetTrack::WidgetTrack(QWidget* parent)
    : QWidget(parent)
{
    Utility::num_tracks++;

    m_track_data = new TrackData();
    m_instrument = new Instrument();

    m_is_muted = false;
    m_is_recording = false;

    QVBoxLayout* control_layout = new QVBoxLayout();

    QTextEdit* label_edit = new QTextEdit();
    label_edit->setMaximumSize(64,48);
    label_edit->setMinimumSize(64,48);

    QPushButton* close_button = new QPushButton(tr("Close"), this);
    close_button->setMaximumSize(64,48);
    close_button->setMinimumSize(64,48);

    mute_button = new QPushButton(tr("Mute"), this);
    mute_button->setMaximumSize(64,48);
    mute_button->setMinimumSize(64,48);

    record_button = new QPushButton(tr("Record"), this);
    record_button->setMaximumSize(64,48);
    record_button->setMinimumSize(64,48);

    control_layout->addWidget(label_edit);
    control_layout->addWidget(close_button);
    control_layout->addWidget(mute_button);
    control_layout->addWidget(record_button);
    control_layout->insertStretch(-1);

    QHBoxLayout* layout = new QHBoxLayout();
    layout->addLayout(control_layout);

    // Matrix control for editing sequence
    matrix_editor = new WidgetMatrixEditor(m_track_data, this);
    matrix_editor->setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));
    layout->addWidget(matrix_editor);
    setLayout(layout);

    // Delete on close
    setAttribute(Qt::WA_DeleteOnClose);

    // Make connections
    connect(close_button, SIGNAL(clicked()), this, SLOT(close()));
    connect(mute_button, SIGNAL(clicked()), this, SLOT(mute_slot()));
    connect(record_button, SIGNAL(clicked()), this, SLOT(record_slot()));
}

float WidgetTrack::get_next_sample()
{

    float result = m_instrument->get_next_sample();
    if (m_is_muted)
    {
        result = 0.0;
    }
    return result;
}

void WidgetTrack::play_beat(unsigned int beat)
{
    // Get the value from the matrix editor.  If it's a start value,
    // then set up the sinusoid(s).
    int note = m_track_data->m_sequence[beat];
    if (note == Utility::NOTE_OFF)
    {
        m_instrument->stop();
    }
    else if (note >= 0)
    {
        m_instrument->play(note);
    }
    matrix_editor->set_beat(beat);
}

void WidgetTrack::stop_playing()
{
    m_instrument->stop();
    // Magic number -1 means don't display timebar for current beat
    matrix_editor->set_beat(-1);
}

void WidgetTrack::closeEvent(QCloseEvent* event)
{
    Utility::get_utility()->lock();
    Utility::num_tracks--;
    event->accept();
    Utility::get_utility()->unlock();
}

void WidgetTrack::record_slot()
{
    // Only let one track record at a time
    if (m_is_recording || (!m_is_recording && !Utility::get_utility()->is_recording()))
    {
        m_is_recording = !m_is_recording;
        // Change background color and interact with utility class
        if (m_is_recording)
        {
            record_button->setAutoFillBackground(true);
            record_button->setStyleSheet("background-color: rgb(128, 0, 0);");
            m_instrument->set_mode(UNIT_SINE_OFF_MODE);
            Utility::get_utility()->lock();
            Utility::get_utility()->start_recording();
            Utility::get_utility()->unlock();
        }
        else
        {
            Utility::get_utility()->lock();
            Utility::get_utility()->stop_recording();
            Utility::get_utility()->unlock();
            m_instrument->get_sample_from_utility();
            m_instrument->do_sms_processing();
            m_instrument->set_mode(UNIT_SINE_SMS_MODE);
            record_button->setStyleSheet(styleSheet());
        }
    }
}

void WidgetTrack::mute_slot()
{
    m_is_muted = !m_is_muted;
    // Change background color if it's muted
    if (m_is_muted)
    {
        mute_button->setAutoFillBackground(true);
        mute_button->setStyleSheet("background-color: rgb(128, 128, 0);");
    }
    else
    {
        mute_button->setStyleSheet(styleSheet());
    }
}

void WidgetTrack::fade_slot()
{
    matrix_editor->fade();
}
